@@ -10,24 +10,21 @@ import decimal
 from decimal import Decimal
 from typing import Sequence, Optional, Mapping, Dict, Union, Any
 
-from aiorpcx.curio import timeout_after, TaskTimeout
+from aiorpcx.curio import timeout_after, TaskTimeout, ignore_after
 import aiohttp
 
 from . import util
 from .bitcoin import COIN
 from .i18n import _
 from .util import (ThreadJob, make_dir, log_exceptions, OldTaskGroup,
-                   make_aiohttp_session, resource_path, EventListener, event_listener, to_decimal)
+                   make_aiohttp_session, resource_path, EventListener, event_listener, to_decimal,
+                   timestamp_to_datetime)
+from .util import NetworkRetryManager
 from .network import Network
 from .simple_config import SimpleConfig
 from .logging import Logger
 
 
-DEFAULT_ENABLED = False
-DEFAULT_CURRENCY = "EUR"
-DEFAULT_EXCHANGE = "CoinGecko"  # default exchange should ideally provide historical rates
-
-
 # See https://en.wikipedia.org/wiki/ISO_4217
 CCY_PRECISIONS = {'BHD': 3, 'BIF': 0, 'BYR': 0, 'CLF': 4, 'CLP': 0,
                   'CVE': 0, 'DJF': 0, 'GNF': 0, 'IQD': 3, 'ISK': 0,
@@ -36,11 +33,12 @@ CCY_PRECISIONS = {'BHD': 3, 'BIF': 0, 'BYR': 0, 'CLF': 4, 'CLP': 0,
                   'RWF': 0, 'TND': 3, 'UGX': 0, 'UYI': 0, 'VND': 0,
                   'VUV': 0, 'XAF': 0, 'XAU': 4, 'XOF': 0, 'XPF': 0,
                   # Cryptocurrencies
-                  'BTC': 8, 'LTC': 8, 'XRP': 6, 'ETH': 18,
+                  'BTC': 8, 'LTC': 6, 'XRP': 4, 'ETH': 8,
                   }
 
-POLL_PERIOD_SPOT_RATE = 150  # approx. every 2.5 minutes, try to refresh spot price
-EXPIRY_SPOT_RATE = 600       # spot price becomes stale after 10 minutes
+SPOT_RATE_REFRESH_TARGET = 150      # approx. every 2.5 minutes, try to refresh spot price
+SPOT_RATE_CLOSE_TO_STALE = 450      # try harder to fetch an update if price is getting old
+SPOT_RATE_EXPIRY = 600              # spot price becomes stale after 10 minutes -> we no longer show/use it
 
 
 class ExchangeBase(Logger):
@@ -88,13 +86,16 @@ class ExchangeBase(Logger):
             self._quotes = await self.get_rates(ccy)
             assert all(isinstance(rate, (Decimal, type(None))) for rate in self._quotes.values()), \
                 f"fx rate must be Decimal, got {self._quotes}"
-            self._quotes_timestamp = time.time()
-            self.logger.info("received fx quotes")
-        except (aiohttp.ClientError, asyncio.TimeoutError) as e:
+        except (aiohttp.ClientError, asyncio.TimeoutError, OSError) as e:
             self.logger.info(f"failed fx quotes: {repr(e)}")
+            self.on_quotes()
         except Exception as e:
             self.logger.exception(f"failed fx quotes: {repr(e)}")
-        self.on_quotes()
+            self.on_quotes()
+        else:
+            self.logger.info("received fx quotes")
+            self._quotes_timestamp = time.time()
+            self.on_quotes(received_new_data=True)
 
     def read_historical_rates(self, ccy: str, cache_dir: str) -> Optional[dict]:
         filename = os.path.join(cache_dir, self.name() + '_'+ ccy)
@@ -121,7 +122,7 @@ class ExchangeBase(Logger):
             self.logger.info(f"requesting fx history for {ccy}")
             h = await self.request_history(ccy)
             self.logger.info(f"received fx history for {ccy}")
-        except (aiohttp.ClientError, asyncio.TimeoutError) as e:
+        except (aiohttp.ClientError, asyncio.TimeoutError, OSError) as e:
             self.logger.info(f"failed fx history: {repr(e)}")
             return
         except Exception as e:
@@ -174,7 +175,7 @@ class ExchangeBase(Logger):
         rate = self._quotes.get(ccy)
         if rate is None:
             return Decimal('NaN')
-        if self._quotes_timestamp + EXPIRY_SPOT_RATE < time.time():
+        if self._quotes_timestamp + SPOT_RATE_EXPIRY < time.time():
             # Our rate is stale. Probably better to return no rate than an incorrect one.
             return Decimal('NaN')
         return Decimal(rate)
@@ -306,7 +307,7 @@ class CoinCap(ExchangeBase):
         # (and history starts on 2017-03-23)
         history = await self.get_json('api.coincap.io',
                                       '/v2/assets/bitcoin/history?interval=d1&limit=2000')
-        return dict([(datetime.utcfromtimestamp(h['time']/1000).strftime('%Y-%m-%d'), str(h['priceUsd']))
+        return dict([(timestamp_to_datetime(h['time']/1000, utc=True).strftime('%Y-%m-%d'), str(h['priceUsd']))
                      for h in history['data']])
 
 
@@ -354,10 +355,27 @@ class CoinGecko(ExchangeBase):
         history = await self.get_json('api.coingecko.com',
                                       '/api/v3/coins/bitcoin/market_chart?vs_currency=%s&days=max' % ccy)
 
-        return dict([(datetime.utcfromtimestamp(h[0]/1000).strftime('%Y-%m-%d'), str(h[1]))
+        return dict([(timestamp_to_datetime(h[0]/1000, utc=True).strftime('%Y-%m-%d'), str(h[1]))
                      for h in history['prices']])
 
 
+class Bit2C(ExchangeBase):
+
+    async def get_rates(self, ccy):
+        json = await self.get_json('bit2c.co.il', '/Exchanges/BtcNis/Ticker.json')
+        return {'ILS': to_decimal(json['ll'])}
+
+    def history_ccys(self):
+        return CURRENCIES[self.name()]
+
+    async def request_history(self, ccy):
+        history = await self.get_json('bit2c.co.il',
+                                      '/Exchanges/BtcNis/KLines?resolution=1D&from=1357034400&to=%s' % int(time.time()))
+
+        return dict([(timestamp_to_datetime(h[0], utc=True).strftime('%Y-%m-%d'), str(h[6]))
+                     for h in history])
+
+
 class CointraderMonitor(ExchangeBase):
 
     async def get_rates(self, ccy):
@@ -387,14 +405,6 @@ class Kraken(ExchangeBase):
                      for k, v in json['result'].items())
 
 
-class LocalBitcoins(ExchangeBase):
-
-    async def get_rates(self, ccy):
-        json = await self.get_json('localbitcoins.com',
-                             '/bitcoinaverage/ticker-all-currencies/')
-        return dict([(r, to_decimal(json[r]['rates']['last'])) for r in json])
-
-
 class MercadoBitcoin(ExchangeBase):
 
     async def get_rates(self, ccy):
@@ -517,10 +527,17 @@ def get_exchanges_by_ccy(history=True):
     return dictinvert(d)
 
 
-class FxThread(ThreadJob, EventListener):
+class FxThread(ThreadJob, EventListener, NetworkRetryManager[str]):
 
     def __init__(self, *, config: SimpleConfig):
         ThreadJob.__init__(self)
+        NetworkRetryManager.__init__(
+            self,
+            max_retry_delay_normal=SPOT_RATE_REFRESH_TARGET,
+            init_retry_delay_normal=SPOT_RATE_REFRESH_TARGET,
+            max_retry_delay_urgent=SPOT_RATE_REFRESH_TARGET,
+            init_retry_delay_urgent=1,
+        )  # note: we poll every 5 seconds for action, so we won't attempt connections more frequently than that.
         self.config = config
         self.register_callbacks()
         self.ccy = self.get_currency()
@@ -535,6 +552,7 @@ class FxThread(ThreadJob, EventListener):
 
     @event_listener
     def on_event_proxy_set(self, *args):
+        self._clear_addr_retry_times()
         self._trigger.set()
 
     @staticmethod
@@ -572,54 +590,66 @@ class FxThread(ThreadJob, EventListener):
 
     async def run(self):
         while True:
-            # every few minutes, refresh spot price
-            try:
-                async with timeout_after(POLL_PERIOD_SPOT_RATE):
-                    await self._trigger.wait()
-                    self._trigger.clear()
-                # we were manually triggered, so get historical rates
-                if self.is_enabled() and self.has_history():
-                    self.exchange.get_historical_rates(self.ccy, self.cache_dir)
-            except TaskTimeout:
-                pass
-            if self.is_enabled():
+            # keep polling and see if we should refresh spot price or historical prices
+            manually_triggered = False
+            async with ignore_after(5):
+                await self._trigger.wait()
+                self._trigger.clear()
+                manually_triggered = True
+            if not self.is_enabled():
+                continue
+            if manually_triggered and self.has_history():  # maybe refresh historical prices
+                self.exchange.get_historical_rates(self.ccy, self.cache_dir)
+            now = time.time()
+            if not manually_triggered and self.exchange._quotes_timestamp + SPOT_RATE_REFRESH_TARGET > now:
+                continue  # last quote still fresh
+            # If the last quote is relatively recent, we poll at fixed time intervals.
+            # Once it gets close to cache expiry, we change to an exponential backoff, to try to get
+            # a quote before it expires. Also, on Android, we might come back from a sleep after a long time,
+            # with the last quote close to expiry or already expired, in that case we go into exponential backoff.
+            is_urgent = self.exchange._quotes_timestamp + SPOT_RATE_CLOSE_TO_STALE < now
+            addr_name = "spot-urgent" if is_urgent else "spot"  # this separates retry-counters
+            if self._can_retry_addr(addr_name, urgent=is_urgent):
+                self._trying_addr_now(addr_name)
+                # refresh spot price
                 await self.exchange.update_safe(self.ccy)
 
-    def is_enabled(self):
-        return bool(self.config.get('use_exchange_rate', DEFAULT_ENABLED))
+    def is_enabled(self) -> bool:
+        return self.config.FX_USE_EXCHANGE_RATE
 
-    def set_enabled(self, b):
-        self.config.set_key('use_exchange_rate', bool(b))
+    def set_enabled(self, b: bool) -> None:
+        self.config.FX_USE_EXCHANGE_RATE = b
         self.trigger_update()
 
     def can_have_history(self):
         return self.is_enabled() and self.ccy in self.exchange.history_ccys()
 
     def has_history(self) -> bool:
-        return self.can_have_history() and bool(self.config.get('history_rates', False))
+        return self.can_have_history() and self.config.FX_HISTORY_RATES
 
     def get_currency(self) -> str:
         '''Use when dynamic fetching is needed'''
-        return self.config.get("currency", DEFAULT_CURRENCY)
+        return self.config.FX_CURRENCY
 
     def config_exchange(self):
-        return self.config.get('use_exchange', DEFAULT_EXCHANGE)
+        return self.config.FX_EXCHANGE
 
     def set_currency(self, ccy: str):
         self.ccy = ccy
-        self.config.set_key('currency', ccy, True)
+        self.config.FX_CURRENCY = ccy
         self.trigger_update()
         self.on_quotes()
 
     def trigger_update(self):
+        self._clear_addr_retry_times()
         loop = util.get_asyncio_loop()
         loop.call_soon_threadsafe(self._trigger.set)
 
     def set_exchange(self, name):
-        class_ = globals().get(name) or globals().get(DEFAULT_EXCHANGE)
+        class_ = globals().get(name) or globals().get(self.config.cv.FX_EXCHANGE.get_default_value())
         self.logger.info(f"using exchange {name}")
         if self.config_exchange() != name:
-            self.config.set_key('use_exchange', name, True)
+            self.config.FX_EXCHANGE = name
         assert issubclass(class_, ExchangeBase), f"unexpected type {class_} for {name}"
         self.exchange = class_(self.on_quotes, self.on_history)  # type: ExchangeBase
         # A new exchange means new fx quotes, initially empty.  Force
@@ -627,7 +657,9 @@ class FxThread(ThreadJob, EventListener):
         self.trigger_update()
         self.exchange.read_historical_rates(self.ccy, self.cache_dir)
 
-    def on_quotes(self):
+    def on_quotes(self, *, received_new_data: bool = False):
+        if received_new_data:
+            self._clear_addr_retry_times()
         util.trigger_callback('on_quotes')
 
     def on_history(self):
@@ -664,13 +696,16 @@ class FxThread(ThreadJob, EventListener):
     def fiat_value(self, satoshis, rate) -> Decimal:
         return Decimal('NaN') if satoshis is None else Decimal(satoshis) / COIN * Decimal(rate)
 
-    def value_str(self, satoshis, rate) -> str:
-        return self.format_fiat(self.fiat_value(satoshis, rate))
+    def value_str(self, satoshis, rate, *, add_thousands_sep: bool = None) -> str:
+        fiat_val = self.fiat_value(satoshis, rate)
+        return self.format_fiat(fiat_val, add_thousands_sep=add_thousands_sep)
 
-    def format_fiat(self, value: Decimal) -> str:
+    def format_fiat(self, value: Decimal, *, add_thousands_sep: bool = None) -> str:
         if value.is_nan():
             return _("No data")
-        return self.ccy_amount_str(value, add_thousands_sep=True)
+        if add_thousands_sep is None:
+            add_thousands_sep = True
+        return self.ccy_amount_str(value, add_thousands_sep=add_thousands_sep)
 
     def history_rate(self, d_t: Optional[datetime]) -> Decimal:
         if d_t is None:
@@ -697,4 +732,4 @@ class FxThread(ThreadJob, EventListener):
         return self.history_rate(date)
 
 
-assert globals().get(DEFAULT_EXCHANGE), f"default exchange {DEFAULT_EXCHANGE} does not exist"
+assert globals().get(SimpleConfig.FX_EXCHANGE.get_default_value()), f"default exchange {SimpleConfig.FX_EXCHANGE.get_default_value()} does not exist"
